import folium
import geopandas as gpd
import pandas.api.types as pdapi
import pandas as pd
from exifScan.creditosfolium.creditosfolium import Creditos
import logging
import pyogrio.errors as pe
import os
from exifScan.generateHTML import generateHTMLSnippets
from bs4 import BeautifulSoup
import shutil


class HTMLWriter:

    def __init__(self, OutputGeopackagePath, htmlFolder, mapFileName, otherMapLayers=[], otherMapLayerNames=None, splitSize=100, linzAPIKey=None, pathPrefix='', windowsPaths=True):
        self.OutputGeopackagePath = OutputGeopackagePath
        self.htmlFolder = htmlFolder
        self.mapFileName = mapFileName
        self.otherMapLayers = otherMapLayers
        self.otherMapLayerNames = otherMapLayerNames
        self.splitSize = splitSize
        self.mapLocation = os.path.join(htmlFolder, mapFileName) if htmlFolder else None
        self.original_color_list = ['#66C5CC', '#F6CF71', '#F89C74', '#DCB0F2', '#87C55F', '#9EB9F3', '#FE88B1', '#C9DB74', '#8BE0A4', '#B497E7', '#D3B484', '#B3B3B3']
        self.dataCategories = ['Historical Aerial Imagery']
        self.linzAPIKey = linzAPIKey
        self.windowsPaths = windowsPaths
        if '|' in pathPrefix:
            self.oldPathPrefix, self.newPathPrefix = pathPrefix.split('|')
        else:
            self.newPathPrefix=pathPrefix
            self.oldPathPrefix=None

        if self.htmlFolder:
            module_dir = os.path.dirname(os.path.abspath(__file__))
            # List of files to copy
            files_to_copy = ['styles.css', 'info.jpg', 'nouislider.min.css', 'nouislider.min.js']

            # Iterate over the array of files and copy each one
            for file_name in files_to_copy:
                src_path = os.path.join(module_dir, file_name)
                dest_path = os.path.join(self.htmlFolder, file_name)
                shutil.copyfile(src_path, dest_path)
                logging.debug(f'Saved {file_name} at {dest_path}')

    def updateHTML(self, info):

        mapGdf = gpd.read_file(self.OutputGeopackagePath, layer='allGroupedData')

        # Clean and organise mapGdf into categories

        mapGdf = mapGdf.drop('fid', axis=1, errors='ignore')
        # convert back to forward slash to make it easy to copy paste. add/replace prefix.
        for col in ['SourceFileDir', 'Metashape Files']:
            if col in mapGdf:
                if self.oldPathPrefix:
                    mapGdf[col] = mapGdf[col].apply(lambda x: x.replace(self.oldPathPrefix, self.newPathPrefix, 1) if x.startswith(self.oldPathPrefix) else x)
                else:
                    mapGdf[col] = self.newPathPrefix + mapGdf[col]
                if self.windowsPaths:
                    mapGdf[col] = mapGdf[col].str.replace('/', '\\')


        logging.info(mapGdf['SourceFileDir']);
        # split data into categories
        categorisedGdfs = {}
        # ensure Geotagged photos is at the top of the list.
        categorisedGdfs['Geotagged photos'] = mapGdf
        if 'Type' in mapGdf.columns:
            for category in self.dataCategories:
                # Filter the GeoDataFrame for the current category
                categoryGdf = mapGdf[mapGdf['Type'] == category]

                # Remove the filtered rows from the original GeoDataFrame
                mapGdf = mapGdf[mapGdf['Type'] != category].copy()
                if len(categoryGdf):
                    categorisedGdfs[category] = categoryGdf

        # remove oversized
        oversized = mapGdf[mapGdf['areaSqkm'] > self.splitSize]
        mapGdf = mapGdf.copy()[mapGdf['areaSqkm'] < self.splitSize]
        if len(mapGdf):
            categorisedGdfs['Geotagged photos'] = mapGdf
        else:
            del categorisedGdfs['Geotagged photos']

        for _, row in oversized.iterrows():
            folder = row['SourceFileDir']
            size = row['areaSqkm']
            logging.warning(f'Folder {folder} is oversized : {size} km2. Please check the folder for unrelated data.')

        # a dict containing the elements to be added to the map
        mapElements = {}
        copyButtonString = ''

        if self.mapLocation:

            # Create a map centered at the centroid of your geodata - centroid not currently used..

            # Calculate the centroid of all geometries
            # centroid = mapGdf.unary_union.centroid
            # logging.debug(f'centroid location: {centroid.y}, {centroid.x}')

            m = folium.Map(location=[-43.58911567661342, 170.00244140625003], zoom_start=7)

            # add Linz basemaps
            if self.linzAPIKey:
                mapTypes = ['aerial']
                for i, mapType in enumerate(mapTypes):
                    urlTemplate = f'https://basemaps.linz.govt.nz/v1/tiles/{mapType}/WebMercatorQuad/{{z}}/{{x}}/{{y}}.webp?api={self.linzAPIKey}'
                    folium.TileLayer(
                        tiles=urlTemplate,
                        attr='Basemap attribution: Sourced from LINZ. CC BY 4.0',
                        name=f'LINZ Data Service - {mapType}',
                        max_zoom=20,
                    ).add_to(m)

            # add Creditos at bottom right.

            content = '<br/> Use the range slider to filter data by Date.<br/>If several features are overlapping, try right-clicking on the feature.'
            if info:
                content = info + content
            Creditos(
                imageurl="info.jpg",
                imagealt="Information Icon",
                tooltip="Information",
                width="36px",
                height="36px",
                expandcontent=content,
            ).add_to(m)

            # add map data.
            # a copy button to copy the file path in the table.
            copyButtonString = generateHTMLSnippets.copyButton()
            dfs = list(categorisedGdfs.values())

            # Find the earliest year
            earliest_year = find_earliest_year(dfs)
            for k, v in categorisedGdfs.items():
                category_id = k.replace(' ', '_')

                addExifLayerToMap(v, k, m, category_id, mapElements)
                copyButtonString += f'makeCellsCopyable("table{category_id}", "SourceFileDir");'
            if self.otherMapLayers:
                for i, layerPath in enumerate(self.otherMapLayers):

                    try:
                        if self.otherMapLayerNames and self.otherMapLayerNames[i]:
                            otherGdf = gpd.read_file(layerPath, layer=self.otherMapLayerNames[i])
                        else:
                            otherGdf = gpd.read_file(layerPath)

                        otherGdf = sanitiseTableGdf(otherGdf)
                        otherGdf = otherGdf.to_crs(epsg=4326)

                        layername = self.otherMapLayerNames[i] if self.otherMapLayerNames else f'Layer {i + 1}'

                        tableDataOther = otherGdf.drop('geometry', axis=1)
                        mapElements[layername] = {}

                        mapElements[layername]['table'] = tableDataOther

                        popupOther = folium.GeoJsonPopup(fields=tableDataOther.columns.tolist(), labels=True, class_name="mystyle", max_width="860px")

                        # shim for sfmData gdb
                        def style_function(feature):
                            year = feature['properties'].get('Photographed_On', None)
                            if year:
                                className = f"year-{year.split('-')[0]} animateVisibility {layername}"
                            else:
                                className = f"animateVisibility {layername}"

                            return {
                                'className': className,
                                'fillOpacity': 0.1
                            }

                        otherGeojson = folium.GeoJson(data=otherGdf, show=False, popup=popupOther, class_name='animateVisibility ' + layername, name=layername, style_function=style_function, highlight_function=lambda feature: {'fillOpacity': 0.5})
                        otherGeojson.add_to(m)
                        mapElements[layername]['geojson'] = otherGeojson

                    except pe.DataSourceError:
                        logging.error(f'{layerPath} does not exist in the file system, and is not recognized as a supported dataset name.')
                    except Exception as e:
                        logging.exception(f'Unexpected error: {e}')

            folium.LayerControl().add_to(m)
            # add geojson css styling to file.
            geojson_styling = '<style>'
            color_list = self.original_color_list.copy()
            for name, value in mapElements.items():
                if 'geojson' in value:
                    geojson_styling += generateHTMLSnippets.cssColors_for_geojson(name, color_list, mapElements)
            geojson_styling += '</style>'
            m.get_root().header.add_child(folium.Element(geojson_styling))

            rightclickJS = generateHTMLSnippets.rightClick(mapElements)
            m.get_root().html.add_child(folium.Element(rightclickJS))

            m.get_root().html.add_child(folium.JavascriptLink('nouislider.min.js'))
            # Link your CSS file to the map
            m.get_root().header.add_child(folium.CssLink('styles.css'))
            m.get_root().header.add_child(folium.CssLink('nouislider.min.css'))

            m.add_css_link(
                "bootstrap_css",
                'https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/5.3.0/css/bootstrap.min.css'
            )
            m.add_js_link(
                "bootstrap_js",
                'https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/5.3.0/js/bootstrap.bundle.min.js'
            )

            # Create the HTML for the button
            bottom_left_ui = '''
            <div class="container custom-container">
                <div class="row align-items-center">
                    <div class="col-2">
                        <button class="btn btn-primary" type="button" data-bs-toggle="offcanvas" data-bs-target="#offcanvasScrolling" aria-controls="offcanvasBottom">Show Table</button>
                    </div>
                    <div class="col-6">
                        <div id="sliderContainer">
                        </div>
                    </div>
                </div>
            </div>
            '''

            # Create a Folium element with the HTML
            element = folium.Element(bottom_left_ui)

            # Add the element to the map
            m.get_root().html.add_child(element)

            m.save(self.mapLocation)

            # Save it to an HTML file

            logging.info(f'Saved new map at {self.mapLocation}')

            # Open the HTML file and parse it
            with open(self.mapLocation, 'r') as f:
                soup = BeautifulSoup(f, 'html.parser')

            new_title = soup.new_tag('title')
            new_title.string = "Map"

            # Add the <title> tag to the <head> section
            soup.head.append(new_title)
            # Add a link to a CSS file in the header
            head = soup.head
            link2 = soup.new_tag('link', rel='stylesheet', href='https://cdn.datatables.net/2.0.2/css/dataTables.bootstrap5.css')
            script3 = soup.new_tag('script', src='https://cdn.datatables.net/2.0.2/js/dataTables.js')
            script4 = soup.new_tag('script', src='https://cdn.datatables.net/2.0.2/js/dataTables.bootstrap5.js')
            head.append(link2)
            head.append(script3)
            head.append(script4)

            body = soup.body

            # Add some HTML content in the body

            # Create navlink, tab and js for the geotagged images table

            navlinks = []
            tabsContent = []
            jsContent = [generateHTMLSnippets.year_filter_slider(earliest_year)]

            # Create navlink, tab and js for other tables

            for key, value in mapElements.items():
                name = key
                el_id = key.replace(' ', '_')

                if 'geojson' in value:
                    geojson_element = value['geojson']
                    head.append(generateHTMLSnippets.zoomOnEl(soup, m.get_name(), f'table{el_id}', geojson_element.get_name()))
                if 'table' in value:
                    table_element = value['table']

                    navlinks.append(generateHTMLSnippets.navLink(el_id, name, name == 'Geotagged photos'))
                    tabsContent.append(generateHTMLSnippets.tabDivs(el_id, name, table_element, name == 'Geotagged photos'))
                    jsContent.append(generateHTMLSnippets.tableJs(el_id))

            # concatenate all the html as strings

            html = generateHTMLSnippets.HTMLTail(navlinks, tabsContent, jsContent, copyButtonString)
            body.append(BeautifulSoup(html, 'html.parser'))
            # Write the changes back to the HTML file
            with open(self.mapLocation, 'w') as f:
                f.write(str(soup))

            logging.debug(f'Added table at {self.mapLocation}')


def sanitiseTableGdf(gdf):
    for col in gdf.columns:
        # If the column is of datetime type
        if pd.api.types.is_datetime64_any_dtype(gdf[col]):
            # Convert the column to string
            gdf[col] = gdf[col].astype(str)
    gdf['index'] = gdf.index

    return gdf


def addExifLayerToMap(gdf, name, m, layer_id, mapElementsDict):
    # Create a DataFrame with the feature data
    data = gdf.drop('geometry', axis=1)
    gdf = gdf.to_crs(epsg=4326)

    # Extract the year and find the earliest year

    categoryTableData = sanitiseTableGdf(data)
    mapElementsDict[name] = {}
    mapElementsDict[name]['table'] = categoryTableData

    # make the SourceFileDir copyable. See rightClick.js for other part of applicable code
    gdf['SourceFileDir'] = gdf['SourceFileDir'] + '<div class="copyText"></div>'

    # make the Metashape Files copyable. See rightClick.js for other part of applicable code
    if 'Metashape Files' in gdf.columns:
        gdf['Metashape Files'] = gdf['Metashape Files'].str.replace('.psx,\n', '.psx<div class="copyText"></div></span><span>')
        gdf['Metashape Files'] = gdf['Metashape Files'].apply(
            lambda x: f'<span>{x}<div class="copyText"></div></span>' if x else x
        )

    # Create a GeoJsonPopup
    popup = folium.GeoJsonPopup(fields=data.columns.tolist(), labels=True, class_name="mystyle", max_width="860px",)

    # Create a GeoJson object with the popup
    mapCopy = gdf.copy()
    mapCopy['index'] = mapCopy.index
    # only show geotagged images layer on load.
    show = (name == 'Geotagged photos')

    def style_function(feature):
        year = feature['properties']['CreateDate'].split('-')[0]
        return {
            'className': f"year-{year} animateVisibility {layer_id}",
            'fillOpacity': 0.1
        }
    geojson = folium.GeoJson(data=mapCopy, name=name, show=show, popup=popup, popup_keep_highlighted=True, style_function=style_function, highlight_function=lambda feature: {'fillOpacity': 0.8})

    mapElementsDict[name]['geojson'] = geojson

    # Add the GeoJson object to the map
    m.add_child(geojson)
    logging.debug(f"{name} added to the map. It has {len(gdf)} features.")

# Function to find the earliest year across multiple DataFrames


def find_earliest_year(dfs):
    earliest_year = float('inf')  # Initialize with a very large number
    for df in dfs:
        temp_dates = pd.to_datetime(df['CreateDate'])
        min_year = temp_dates.dt.year.min()
        if min_year < earliest_year:
            earliest_year = min_year
    return earliest_year


# Example event code- poorly documented so leaving it here.
# from folium.utilities import JsCode
# from folium.elements import EventHandler

# highlight = JsCode(
#     """
#    function highlight(e) {
#        console.log(e)
#    }
# """
# )

# reset = JsCode(
#     """
#    function reset(e) {
#       e.target.setStyle({ color: e.target.original_color });
#    }
# """
# )

# geojson.add_child(EventHandler("load", highlight))
# geojson.add_child(EventHandler("mouseout", reset))



import pytest
from unittest.mock import patch
from exifScan.watchDirectory import watchFolder
import time
import tempfile
import os
import logging


@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir


@pytest.fixture
def temp_file():
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        yield tmpfile.name
    os.remove(tmpfile.name)


@pytest.fixture
def args_dict(temp_dir):
    return {
        'ignoreMetashape': True,
        'ignoreDirs': [],
        'readExifFromOutputFile': None,
        'pollingInterval': 10,
        'persistObserver': True,
        'WorkDirectory': temp_dir,
        'GeopackageDirectory': temp_dir,
        'OutputGeopackage': 'output.gpkg',
        'timerInterval': 0.5
    }


@patch('exifScan.watchDirectory.ExifScanner')
def test_watch_folder_initial_scan(mock_scanner, args_dict, temp_dir, temp_file):

    # Start the watch process in a separate thread
    import threading
    watch_thread = threading.Thread(target=watchFolder, args=(temp_dir, args_dict))
    watch_thread.start()
    watch_thread.join()
    mock_scanner.assert_called_with(**{'ignoreMetashape': True, 'ignoreDirs': [], 'readExifFromOutputFile': None, 'WorkDirectory': temp_dir, 'GeopackageDirectory': temp_dir, 'OutputGeopackage': 'output.gpkg', 'recursive': True})


@patch('exifScan.watchDirectory.ExifScanner')
def test_watch_folder_reuse_snapshot(mock_scanner, args_dict, temp_dir):

    watchFolder(temp_dir, args_dict)

    mock_scanner.assert_called_with(**{'ignoreMetashape': True, 'ignoreDirs': [], 'readExifFromOutputFile': None, 'WorkDirectory': temp_dir, 'GeopackageDirectory': temp_dir, 'OutputGeopackage': 'output.gpkg', 'recursive': True})
    mock_scanner.return_value.scan_for_photos.assert_called_with(temp_dir)


@patch('exifScan.watchDirectory.os.path.exists')
@patch('exifScan.watchDirectory.ExifScanner')
@patch('exifScan.watchDirectory.FolderUpdateHandler')
@patch('exifScan.watchDirectory.PersistentPollingObserver')
def test_watch_folder_output_geopackage_exists(mock_observer, mock_handler, mock_scanner, mock_exists, args_dict, temp_dir):
    mock_scanner.return_value.scan_for_photos.return_value = temp_dir
    mock_exists.return_value = True
    mock_observer_instance = mock_observer.return_value
    mock_observer_instance.previous_snapshot_exists = False

    with patch('builtins.input', return_value='y'):
        import threading
        stop_event = threading.Event()
        watch_thread = threading.Thread(target=watchFolder, args=(temp_dir, args_dict, stop_event))
        watch_thread.start()
        stop_event.set()
    mock_scanner.assert_called_with(**{'ignoreMetashape': True, 'ignoreDirs': [], 'readExifFromOutputFile': None, 'WorkDirectory': temp_dir, 'GeopackageDirectory': temp_dir, 'OutputGeopackage': 'output.gpkg', 'recursive': True})
    mock_handler.assert_called_with(['.files', 'auto-project'], mock_scanner.return_value, timer_interval=0.5)
    mock_observer_instance.schedule.assert_called_once()
    mock_observer_instance.start.assert_called_once()


@patch('exifScan.watchDirectory.ExifScanner')
def test_watch_folder_file_modification(mock_scanner, args_dict, temp_dir, temp_file, caplog):
    mock_scanner.return_value.scan_for_photos.return_value = temp_dir
    with caplog.at_level(logging.DEBUG):
        # Start the watch process in a separate thread
        import threading
        stop_event = threading.Event()
        watch_thread = threading.Thread(target=watchFolder, args=(temp_dir, args_dict, stop_event))
        watch_thread.start()

        # Give the observer some time to start
        time.sleep(2)

        # Simulate a file modification
        test_file_path = os.path.join(temp_dir, 'test_file.txt')
        with open(test_file_path, 'w') as f:
            f.write('Test content')
        # Give the observer some time to detect the change
        time.sleep(10)

        # Assert that the event handler's methods were called

        # Clean up
        stop_event.set()
        watch_thread.join()
        os.remove(test_file_path)

        mock_scanner.return_value.scan_for_photos.assert_called_with(temp_dir)
        assert mock_scanner.return_value.scan_for_photos.call_count == 2



import geopandas as gpd
import os
import pytest
from exifScan.generateHTML.createHTML import HTMLWriter  # Replace 'exifScan.generateHTML.createHTML' with the actual module name
import tempfile
from shapely.geometry import Point
import logging
from bs4 import BeautifulSoup
import shutil


@pytest.fixture
def setup_test_environment():
    # Create a temporary directory
    test_dir = tempfile.mkdtemp()
    yield test_dir
    # Cleanup after tests
    shutil.rmtree(test_dir)

def test_html_writer_initialization(setup_test_environment):
    # Test data
    test_dir = setup_test_environment
    OutputGeopackagePath = "path/to/geopackage"
    htmlFolder = test_dir
    mapFileName = "map.html"
    otherMapLayers = ["layer1", "layer2"]
    otherMapLayerNames = ["Layer 1", "Layer 2"]
    splitSize = 10

    # Create an instance of HTMLWriter
    writer = HTMLWriter(OutputGeopackagePath, htmlFolder, mapFileName, otherMapLayers, otherMapLayerNames, splitSize)

    # Assertions
    assert writer.OutputGeopackagePath == OutputGeopackagePath
    assert writer.htmlFolder == htmlFolder
    assert writer.mapFileName == mapFileName
    assert writer.otherMapLayers == otherMapLayers
    assert writer.otherMapLayerNames == otherMapLayerNames
    assert writer.splitSize == splitSize
    assert writer.mapLocation == os.path.join(htmlFolder, mapFileName)
    assert writer.original_color_list == ['#66C5CC', '#F6CF71', '#F89C74', '#DCB0F2', '#87C55F', '#9EB9F3', '#FE88B1', '#C9DB74', '#8BE0A4', '#B497E7', '#D3B484', '#B3B3B3']

def test_html_writer_file_copy(setup_test_environment):
    # Test data
    test_dir = setup_test_environment
    OutputGeopackagePath = "path/to/geopackage"
    htmlFolder = test_dir
    mapFileName = "map.html"
    
    # Create dummy files to copy in the module directory
    module_dir = os.path.dirname(os.path.abspath(__file__))
    files_to_copy = ['styles.css', 'info.jpg', 'nouislider.min.css', 'nouislider.min.js']
    
    for file_name in files_to_copy:
        with open(os.path.join(module_dir, file_name), 'w') as f:
            f.write("dummy content")

    # Create an instance of HTMLWriter
    writer = HTMLWriter(OutputGeopackagePath, htmlFolder, mapFileName)

    # Check if files are copied correctly
    for file_name in files_to_copy:
        dest_path = os.path.join(htmlFolder, file_name)
        assert os.path.exists(dest_path), f"{file_name} was not copied to {dest_path}"

@pytest.fixture
def html_writer(setup_test_environment):
    test_dir = setup_test_environment
    layer_1_dir = os.path.join(test_dir, "layer_1.shp")
    layer_2_dir = os.path.join(test_dir, "layer_2.shp")

    return HTMLWriter(
        OutputGeopackagePath=os.path.join(test_dir, "test.gpkg"),
        htmlFolder=test_dir,
        mapFileName="map.html",
        otherMapLayers=[layer_1_dir, layer_2_dir],
        splitSize=10,
        pathPrefix='/old/path|/new/path',
        windowsPaths=True
    )

    
def test_updateHTML(html_writer, setup_test_environment, caplog):
    test_dir = setup_test_environment
    # Create a mock GeoDataFrame for existing data
    test_data = {
        'SourceFileDir': ['/old/path/dir1', '\old\path\dir2', '/old/path/dir3'],
        'Make': ['Canon', 'Nikon', 'Sony'],
        'Model': ['EOS', 'D850', 'A7'],
        'areaSqkm': [1, 2, 3],
        'CreateDate': ['1990-03-03', '1991-06-27', '1992-12-12']
    }

    # Add a geometry column with Point geometries
    geometry = [Point(1, 2), Point(2, 3), Point(3, 4)]
    test_gdf = gpd.GeoDataFrame(test_data, index=[1, 2, 3], crs="EPSG:4326", geometry=geometry)
    test_gdf.to_file(os.path.join(test_dir, "test.gpkg"), layer='allGroupedData', driver="GPKG")
    test_gdf.to_file(html_writer.otherMapLayers[0], driver="ESRI Shapefile")
    test_gdf.to_file(html_writer.otherMapLayers[1], driver="ESRI Shapefile")
    with caplog.at_level(logging.DEBUG):
        html_writer.updateHTML(None)

    assert os.path.exists(os.path.join(test_dir, "map.html"))
    with open(os.path.join(test_dir, "map.html"), 'r') as file:
        map_html_content = file.read()
    assert "<html>" in map_html_content  # Basic check to see if it's an HTML file
    assert map_html_content.count("<html>") == 1, "There should be exactly one occurrence of '<html>'"
    assert map_html_content.count("</html>") == 1, "There should be exactly one occurrence of '</html>'"
    assert "Layer 1" in map_html_content  # Check if the layer name is included
    soup = BeautifulSoup(map_html_content, 'html.parser')
    # Example assertions
    assert soup.title.string == "Map"
    # Check for the folium map div
    folium_map = soup.find('div', class_='folium-map')
    assert folium_map is not None

    # Check for the offcanvas div
    offcanvas = soup.find('div', id='offcanvasScrolling')
    assert offcanvas is not None
    assert offcanvas['aria-labelledby'] == 'offcanvasScrollingLabel'
    assert offcanvas['class'] == ['offcanvas', 'offcanvas-bottom']
    assert offcanvas['data-bs-backdrop'] == 'false'
    assert offcanvas['data-bs-scroll'] == 'true'
    assert offcanvas['tabindex'] == '-1'

    # Check for the offcanvas header
    offcanvas_header = offcanvas.find('div', class_='offcanvas-header')
    assert offcanvas_header is not None

    # Check for the offcanvas body
    offcanvas_body = offcanvas.find('div', class_='offcanvas-body')
    assert offcanvas_body is not None

    # Check for the nav tabs
    nav_tabs = offcanvas_body.find('div', class_='nav nav-tabs')
    assert nav_tabs is not None

    # Check for the tab content
    tab_content = offcanvas_body.find('div', class_='tab-content')
    assert tab_content is not None

    # Check for the table
    table = tab_content.find('table', id='tableGeotagged_photos')
    assert table is not None
    assert table['class'] == ['dataframe', 'table', 'table-striped', 'responsive']

    # Check for table headers
    headers = table.find_all('th')
    assert len(headers) == 10
    assert headers[1].string == 'SourceFileDir'
    assert headers[2].string == 'Make'
    assert headers[3].string == 'Model'
    assert headers[4].string == 'areaSqkm'
    assert headers[5].string == 'CreateDate'
    assert headers[6].string == 'index'
    # Each row had a th for some reason?
    assert headers[7].string == '0'
    assert headers[8].string == '1'
    assert headers[9].string == '2'

    # Check for table rows
    rows = table.find_all('tr')
    assert len(rows) == 4  # Including header row
    assert rows[1].find_all('td')[0].string == '\\new\\path\\dir1'
    assert rows[1].find_all('td')[1].string == 'Canon'
    assert rows[1].find_all('td')[2].string == 'EOS'
    assert rows[1].find_all('td')[3].string == '1'
    assert rows[1].find_all('td')[4].string == '1990-03-03'
    assert rows[1].find_all('td')[5].string == '0'
    assert rows[2].find_all('td')[0].string == '\\old\\path\\dir2'
    assert rows[2].find_all('td')[1].string == 'Nikon'
    assert rows[2].find_all('td')[2].string == 'D850'
    assert rows[2].find_all('td')[3].string == '2'
    assert rows[2].find_all('td')[4].string == '1991-06-27'
    assert rows[2].find_all('td')[5].string == '1'
    assert rows[3].find_all('td')[0].string == '\\new\\path\\dir3'
    assert rows[3].find_all('td')[1].string == 'Sony'
    assert rows[3].find_all('td')[2].string == 'A7'
    assert rows[3].find_all('td')[3].string == '3'
    assert rows[3].find_all('td')[4].string == '1992-12-12'
    assert rows[3].find_all('td')[5].string == '2'

    # Check Scripts
    # Extract the script content
    script_content_1 = soup.findAll('script')[9].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableGeotagged_photos' );" in script_content_1
    # Extract the script content
    script_content_2 = soup.findAll('script')[10].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableLayer_1' );" in script_content_2
    script_content_3 = soup.findAll('script')[11].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableLayer_2' );" in script_content_3
    # Check if the script content contains the expected values
    script_content_4 = soup.findAll('script')[12].string
    assert 'Geotagged photos": "#8BE0A4"' in script_content_4
    assert '"Layer 1": "#F6CF71"' in script_content_4
    assert '"Layer 2": "#C9DB74"' in script_content_4
    script_content_5 = soup.findAll('script')[14].string
    assert "('#tableGeotagged_photos').DataTable({responsive:true});" in script_content_5
    assert "('#tableLayer_1').DataTable({responsive:true});" in script_content_5
    assert "('#tableLayer_2').DataTable({responsive:true});" in script_content_5
    script_content_6 = soup.findAll('script')[15].string
    assert 'makeCellsCopyable("tableGeotagged_photos", "SourceFileDir");' in script_content_6

    test_gdf['areaSqkm'] = [1, 2, 30]
    test_gdf.to_file(os.path.join(test_dir, "test.gpkg"), layer='allGroupedData')
    with caplog.at_level(logging.DEBUG):
        html_writer.updateHTML(None)
    with open(os.path.join(test_dir, "map.html"), 'r') as file:
        map_html_content = file.read()
    soup = BeautifulSoup(map_html_content, 'html.parser')
    offcanvas = soup.find('div', id='offcanvasScrolling')
    # Check for the offcanvas div
    offcanvas = soup.find('div', id='offcanvasScrolling')
    assert offcanvas is not None
    assert offcanvas['aria-labelledby'] == 'offcanvasScrollingLabel'
    assert offcanvas['class'] == ['offcanvas', 'offcanvas-bottom']
    assert offcanvas['data-bs-backdrop'] == 'false'
    assert offcanvas['data-bs-scroll'] == 'true'
    assert offcanvas['tabindex'] == '-1'

    # Check for the offcanvas header
    offcanvas_header = offcanvas.find('div', class_='offcanvas-header')
    assert offcanvas_header is not None

    # Check for the offcanvas body
    offcanvas_body = offcanvas.find('div', class_='offcanvas-body')
    assert offcanvas_body is not None

    # Check for the nav tabs
    nav_tabs = offcanvas_body.find('div', class_='nav nav-tabs')
    assert nav_tabs is not None

    # Check for the tab content
    tab_content = offcanvas_body.find('div', class_='tab-content')
    assert tab_content is not None

    # Check for the table
    table = tab_content.find('table', id='tableGeotagged_photos')
    assert table is not None
    assert table['class'] == ['dataframe', 'table', 'table-striped', 'responsive']

    # Check for table headers
    headers = table.find_all('th')

    assert len(headers) == 9
    assert headers[1].string == 'SourceFileDir'
    assert headers[2].string == 'Make'
    assert headers[3].string == 'Model'
    assert headers[4].string == 'areaSqkm'
    assert headers[5].string == 'CreateDate'
    assert headers[6].string == 'index'
    # Each row had a th for some reason?
    assert headers[7].string == '0'
    assert headers[8].string == '1'

    # Check for table rows
    rows = table.find_all('tr')
    assert len(rows) == 3  # Including header row
    assert rows[1].find_all('td')[0].string == '\\new\\path\\dir1'
    assert rows[1].find_all('td')[1].string == 'Canon'
    assert rows[1].find_all('td')[2].string == 'EOS'
    assert rows[1].find_all('td')[3].string == '1'
    assert rows[1].find_all('td')[4].string == '1990-03-03'
    assert rows[1].find_all('td')[5].string == '0'
    assert rows[2].find_all('td')[0].string == '\\old\\path\\dir2'
    assert rows[2].find_all('td')[1].string == 'Nikon'
    assert rows[2].find_all('td')[2].string == 'D850'
    assert rows[2].find_all('td')[3].string == '2'
    assert rows[2].find_all('td')[4].string == '1991-06-27'
    assert rows[2].find_all('td')[5].string == '1'

    test_gdf['Type'] = [None, None, 'Historical Aerial Imagery']
    test_gdf.to_file(os.path.join(test_dir, "test.gpkg"), layer='allGroupedData')
    with caplog.at_level(logging.DEBUG):
        html_writer.updateHTML(None)
    with open(os.path.join(test_dir, "map.html"), 'r') as file:
        map_html_content = file.read()
    soup = BeautifulSoup(map_html_content, 'html.parser')
    offcanvas = soup.find('div', id='offcanvasScrolling')
    # Check for the offcanvas div
    offcanvas = soup.find('div', id='offcanvasScrolling')
    assert offcanvas is not None
    assert offcanvas['aria-labelledby'] == 'offcanvasScrollingLabel'
    assert offcanvas['class'] == ['offcanvas', 'offcanvas-bottom']
    assert offcanvas['data-bs-backdrop'] == 'false'
    assert offcanvas['data-bs-scroll'] == 'true'
    assert offcanvas['tabindex'] == '-1'

    # Check for the offcanvas header
    offcanvas_header = offcanvas.find('div', class_='offcanvas-header')
    assert offcanvas_header is not None

    # Check for the offcanvas body
    offcanvas_body = offcanvas.find('div', class_='offcanvas-body')
    assert offcanvas_body is not None

    # Check for the nav tabs
    nav_tabs = offcanvas_body.find('div', class_='nav nav-tabs')
    assert nav_tabs is not None

    # Check for the tab content
    tab_content = offcanvas_body.find('div', class_='tab-content')
    assert tab_content is not None

    # Check for the table
    table = tab_content.find('table', id='tableGeotagged_photos')
    assert table is not None
    assert table['class'] == ['dataframe', 'table', 'table-striped', 'responsive']

    # Check for table headers
    headers = table.find_all('th')

    assert len(headers) == 10
    assert headers[1].string == 'SourceFileDir'
    assert headers[2].string == 'Make'
    assert headers[3].string == 'Model'
    assert headers[4].string == 'areaSqkm'
    assert headers[5].string == 'CreateDate'
    assert headers[6].string == 'Type'
    assert headers[7].string == 'index'
    # Each row had a th for some reason?
    assert headers[8].string == '0'
    assert headers[9].string == '1'

    # Check for table rows
    rows = table.find_all('tr')
    assert len(rows) == 3  # Including header row
    assert rows[1].find_all('td')[0].string == '\\new\\path\\dir1'
    assert rows[1].find_all('td')[1].string == 'Canon'
    assert rows[1].find_all('td')[2].string == 'EOS'
    assert rows[1].find_all('td')[3].string == '1'
    assert rows[1].find_all('td')[4].string == '1990-03-03'
    assert rows[1].find_all('td')[5].string == 'None'
    assert rows[1].find_all('td')[6].string == '0'
    assert rows[2].find_all('td')[0].string == '\\old\\path\\dir2'
    assert rows[2].find_all('td')[1].string == 'Nikon'
    assert rows[2].find_all('td')[2].string == 'D850'
    assert rows[2].find_all('td')[3].string == '2'
    assert rows[2].find_all('td')[4].string == '1991-06-27'
    assert rows[2].find_all('td')[5].string == 'None'
    assert rows[2].find_all('td')[6].string == '1'

    # Check for the table
    table = tab_content.find('table', id='tableHistorical_Aerial_Imagery')
    assert table is not None
    assert table['class'] == ['dataframe', 'table', 'table-striped', 'responsive']

    # Check for table headers
    headers = table.find_all('th')

    assert len(headers) == 9
    assert headers[1].string == 'SourceFileDir'
    assert headers[2].string == 'Make'
    assert headers[3].string == 'Model'
    assert headers[4].string == 'areaSqkm'
    assert headers[5].string == 'CreateDate'
    assert headers[6].string == 'Type'
    # Each row had a th for some reason?
    assert headers[7].string == 'index'
    assert headers[8].string == '2'

    # Check for table rows
    rows = table.find_all('tr')
    assert len(rows) == 2  # Including header row
    assert rows[1].find_all('td')[0].string == '\\new\\path\\dir3'
    assert rows[1].find_all('td')[1].string == 'Sony'
    assert rows[1].find_all('td')[2].string == 'A7'
    assert rows[1].find_all('td')[3].string == '30'
    assert rows[1].find_all('td')[4].string == '1992-12-12'
    assert rows[1].find_all('td')[5].string == 'Historical Aerial Imagery'
    assert rows[1].find_all('td')[6].string == '2'
    # assert not updated_gdf.empty
    # assert 'geometry' in updated_gdf.columns
    # Assuming updateHTML splits data into multiple files if it exceeds splitSize
    # Check if multiple files are created or if the data is split correctly
    # Check Scripts
    # Extract the script content
    script_content_1 = soup.findAll('script')[9].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableGeotagged_photos' );" in script_content_1
    # Extract the script content
    script_content_1a = soup.findAll('script')[10].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableHistorical_Aerial_Imagery' );" in script_content_1a
    script_content_2 = soup.findAll('script')[11].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableLayer_1' );" in script_content_2
    script_content_3 = soup.findAll('script')[12].string
    # Check if the script content contains the expected values
    assert "const table = document.getElementById( 'tableLayer_2' );" in script_content_3
    # Check if the script content contains the expected values
    script_content_4 = soup.findAll('script')[13].string
    assert 'Geotagged photos": "#8BE0A4"' in script_content_4
    assert '"Historical Aerial Imagery": "#B497E7"' in script_content_4
    assert '"Layer 1": "#B3B3B3"' in script_content_4
    assert '"Layer 2": "#C9DB74"' in script_content_4
    script_content_5 = soup.findAll('script')[15].string
    assert "('#tableGeotagged_photos').DataTable({responsive:true});" in script_content_5
    assert "('#tableHistorical_Aerial_Imagery').DataTable({responsive:true});" in script_content_5
    assert "('#tableLayer_1').DataTable({responsive:true});" in script_content_5
    assert "('#tableLayer_2').DataTable({responsive:true});" in script_content_5
    script_content_6 = soup.findAll('script')[16].string
    assert 'makeCellsCopyable("tableGeotagged_photos", "SourceFileDir");' in script_content_6
    # read number of rows in data.
    try:
        html_writer.updateHTML(None)
    except Exception as e:
        assert False, f"updateHTML raised an exception: {e}"



import pandas as pd
import numpy as np
import logging
import geopandas as gpd
from datetime import datetime as dt
import os


class GeoDataProcessor:
    def __init__(self, WorkDirectory, xyTagNames, zTagName, groupByList):
        self.WorkDirectory = WorkDirectory
        self.xyTagNames = xyTagNames
        self.zTagName = zTagName
        self.groupByList = groupByList
        self.duplicateColumns = [*groupByList, 'SourceFileName', 'CreateDateTime', 'FileSize']

    def process(self, exifDf):
        x_coords = self.xyTagNames[0]
        y_coords = self.xyTagNames[1]

        if not isinstance(exifDf, pd.DataFrame):
            logging.error("The provided argument is not a DataFrame.")
            # return pd.DataFrame(), pd.DataFrame(), pd.DataFrame()

        if all(item in exifDf.columns for item in self.xyTagNames):
            for tag in self.xyTagNames:
                exifDf[tag] = pd.to_numeric(exifDf[tag], errors='coerce')
                exifDf[tag] = exifDf[tag].where(exifDf[tag] != 0, np.nan)

            valid_df = exifDf.copy().dropna(subset=self.xyTagNames)
            invalid_df = exifDf.copy()[exifDf[self.xyTagNames].isnull().any(axis=1)]
        else:
            invalid_df = exifDf.copy()
            valid_df = pd.DataFrame(columns=exifDf.columns)

        if len(invalid_df):
            logging.warning(f"{len(invalid_df)} images are not geolocated.")
        logging.info(f"{len(valid_df)} images are geolocated.")

        if not valid_df.empty:
            logging.debug('Handling geolocated images')
            if self.zTagName not in valid_df.columns:
                valid_df[self.zTagName] = 0
                logging.warning(f"Added {self.zTagName} column with default value 0")

            valid_df[self.zTagName] = pd.to_numeric(valid_df[self.zTagName], errors='coerce')
            valid_df[self.zTagName] = valid_df[self.zTagName].fillna(0)

            logging.info("Cleaning up table. Changed altitude NA values to 0")

            applyBounds = True

            if applyBounds:
                ymin, ymax, xmin, xmax = -180, 180, -90, 90

                if ymin > ymax:
                    dataBounded1 = valid_df[valid_df[x_coords] > ymin]
                    dataBounded2 = valid_df[valid_df[x_coords] < ymax]
                    valid_df = pd.concat([dataBounded1, dataBounded2], ignore_index=True)
                else:
                    valid_df = valid_df[(valid_df[x_coords] > ymin) & (valid_df[x_coords] < ymax)]
                    valid_df = valid_df[(valid_df[y_coords] > xmin) & (valid_df[y_coords] < xmax)]

                merged = valid_df.merge(valid_df, how='outer', indicator=True)
                outOfBounds = merged[merged['_merge'] == 'right_only']
                logging.info(f"Removed {len(outOfBounds)} rows that were out of bounds.")

                logging.info(f"Cleaning up table. Data bounded by {ymin} and {ymax} Longitude and by {xmin} and {xmax} Latitude.")

                if valid_df.empty:
                    logging.info('All data was filtered out. Either no NZ locations are included, or the exif data were not valid values.')
            else:
                logging.info('Unbounded data.')

            setDfColumns(valid_df)
            self.getDuplicates(valid_df)

            logging.info("Creating point GeoDataFrame")
            pointsDf = gpd.GeoDataFrame(data=valid_df, geometry=gpd.points_from_xy(valid_df[x_coords], valid_df[y_coords], valid_df[self.zTagName]), crs="EPSG:4326")
            logging.info(f"Points created. It has {len(pointsDf)} features.")

            groupByListGL = [*self.groupByList, 'SourceFileDir']
            groupByListGL = [col for col in groupByListGL if col in pointsDf.columns]

            aggfuncs = {
                'SourceFileName': 'count',
            }

            polygonsDf = pointsDf.dissolve(by=groupByListGL, aggfunc=aggfuncs, as_index=False)
            polygonsDf.rename(columns={'SourceFileName': 'Image Count'}, inplace=True)
            polygonsDf = polygonsDf[[*groupByListGL, 'Image Count', 'geometry']]
            polygonsDf['geometry'] = polygonsDf.convex_hull
            logging.info(f'Grouped points into polygons by {", ".join(groupByListGL)}. It has {len(polygonsDf)} features.')
        else:
            pointsDf = polygonsDf = gpd.GeoDataFrame()

        if not invalid_df.empty:
            logging.debug('Handling non-geolocated images')
            setDfColumns(invalid_df)
            existing_columns = [col for col in self.groupByList if col in invalid_df.columns]
            self.getDuplicates(invalid_df)
            groupByListNGL = [*existing_columns, 'SourceFileDir']
            try:
                nonGeoDf = invalid_df.groupby(groupByListNGL).agg({
                    'SourceFileName': list,
                }).reset_index()
                nonGeoDf['Image Count'] = nonGeoDf['SourceFileName'].str.len()
            except Exception as e:
                logging.error(f"Error while aggregating data : {e}")
                nonGeoDf = pd.DataFrame()
        else:
            nonGeoDf = pd.DataFrame()

        return pointsDf, polygonsDf, nonGeoDf

    def getDuplicates(self, df):

        # get duplicates
        existing_columns = [col for col in self.duplicateColumns if col in df.columns]

        duplicateRows = df.copy()[df.duplicated(existing_columns, keep=False)]

        if (hasattr(df, '_geometry_column_name')):
            identicalGeometryDf = gpd.GeoDataFrame()
            for _, geom in enumerate(duplicateRows.geometry):
                # Get rows where geometry is identical to the current row
                identicalRows = duplicateRows[duplicateRows.geometry.apply(lambda x: x.equals(geom))]

                # Append identical rows to the result DataFrame
                identicalGeometryDf = identicalGeometryDf.append(identicalRows)
            duplicateRows = identicalGeometryDf

        if not duplicateRows.empty:

            logging.warning(f'Found {len(duplicateRows)} duplicate rows (~ {int(df["FileSize"].sum() / 1073741824)} GB).')

            if 'CreateDateTime' in duplicateRows:
                # Convert to datetime with a specified format
                duplicateRows['CreateDateTime'] = pd.to_datetime(duplicateRows['CreateDateTime'], format='%Y:%m:%d %H:%M:%S', errors='coerce')
                # Sort the DataFrame
                duplicateRows = duplicateRows.sort_values(by='CreateDateTime')

            timestamp = dt.now().strftime('%Y%m%d%H%M%S')

            # Create a unique filename using the timestamp
            filename = f'duplicates_{timestamp}.csv'
            filepath = os.path.join(self.WorkDirectory, filename)
            # Write the DataFrame to a CSV file with the unique filename
            try:
                duplicateRows.to_csv(filepath, index=False)
                logging.debug(f'Duplicate rows written to : {filepath} .')

            except Exception as e:
                logging.error(f'Error when writing to {filename} :{e}')


def setDfColumns(df):
    # set up fileName and directory columns
    df['SourceFileDir'] = df['SourceFile'].str.rpartition('/')[0]
    df['SourceFileName'] = df['SourceFile'].str.rpartition('/')[2]
    if 'CreateDate' in df.columns:
        # Strip everything after the seconds value
        df['CreateDate'] = df['CreateDate'].str.slice(0, 10)
        try:
            df['CreateDate'] = pd.to_datetime(df['CreateDate'], format='%Y:%m:%d').dt.date
        except ValueError as e:
            logging.error(f'Unable to parse datetime, will coerce values. Please check source of error: {e}')
            df['CreateDate'] = pd.to_datetime(df['CreateDate'], format='%Y:%m:%d', errors='coerce').dt.date



import subprocess
import logging
import time
import sys


class KartWriter:
    def __init__(self, kart_directory, output_geopackage_path, kart_remote=None, min_delay_minutes=0):
        self.KartDirectory = kart_directory
        self.OutputGeopackagePath = output_geopackage_path
        self.KartRemote = kart_remote
        self.MinDelaySeconds = min_delay_minutes * 60
        self.LastRunTime = 0
        self.kart_installed = self.is_kart_installed()

    def is_kart_installed(self):
        try:
            result = subprocess.run(["kart", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if result.returncode == 0:
                logging.debug("Kart installed")
                return True

        except FileNotFoundError:
            logging.error('Kart is not installed, update will not be pushed to kart.')
            while True:
                confirmation = input("Are you sure you want to proceed? (y/n): ")
                if confirmation.lower() == 'y':
                    break
                elif confirmation.lower() == 'n':
                    logging.warning("Aborting script.")
                    sys.exit()
                else:
                    logging.error("Invalid input. Please enter 'y' or 'n'.")
        return False

    def import_to_kart(self):
        if not self.kart_installed:
            logging.error('Kart not installed.')
        current_time = time.time()
        if current_time - self.LastRunTime < self.MinDelaySeconds:
            logging.info(f"Skipping import to Kart as last write was less than {self.MinDelaySeconds / 60} minutes ago (the specified minimum delay).")
            return

        if self.KartDirectory:
            kartCmd = [
                "kart",
                "import",
                self.OutputGeopackagePath,
                "--all-tables",
                "--replace-existing"
            ]
            try:
                result = subprocess.run(kartCmd, cwd=self.KartDirectory)
                if result.returncode == 41:
                    logging.warning("Kart repository has not been initialised. Initialising now...")
                    init_process = subprocess.run(['kart', 'init', '--import', self.OutputGeopackagePath], cwd=self.KartDirectory)

                    if init_process.returncode == 0:
                        logging.info("Kart repository initialised successfully.")
                    else:
                        logging.error(f"Failed to initialise Kart repository. Return code: {init_process.returncode}")
                elif result.returncode == 44:
                    logging.info("Kart repository recorded no changes.")
                elif result.returncode == 0:
                    logging.info("Kart repository imported successfully.")

                    if self.KartRemote:
                        remoteResult = subprocess.run(['kart', 'push'], cwd=self.KartDirectory)

                        if remoteResult.returncode == 128:
                            logging.warning("Kart remote repository has not been set. Setting now...")
                            remoteAddResult = subprocess.run(['kart', 'remote', 'add', 'origin', self.KartRemote], cwd=self.KartDirectory)
                            pushResult = subprocess.run(['kart', 'push', '--set-upstream', 'origin', 'main'], cwd=self.KartDirectory)
                            if remoteAddResult.returncode == 0 and pushResult.returncode == 0:
                                logging.info(f"Kart {self.KartRemote} remote repository set successfully.")
                            else:
                                logging.error(f"Failed to set Kart remote repository.\nRemote add return code: {remoteAddResult.returncode}\n Push result: {pushResult.returncode}")
                        elif remoteResult.returncode == 0:
                            logging.info('Kart repository pushed successfully.')
                        else:
                            logging.error(f'Kart remote failed with error code: {remoteResult.returncode}')
                else:
                    logging.error(f'Kart import failed with error code: {result.returncode}')

                self.LastRunTime = current_time

            except NotADirectoryError:
                logging.error(f'Kart directory doesn\'t exist: {self.KartDirectory}. Data will not be backed up to a repository.')

# Example usage:
# kart_writer = KartWriter("/path/to/kart_directory", "/path/to/output_geopackage.gpkg", "https://remote.repo.url", min_delay_minutes=30)
# kart_writer.import_to_kart()



import os
import pickle
import threading
import time
import pytest
import tempfile
from watchdog.events import FileCreatedEvent
from watchdog.observers.api import EventQueue
from persistentObserver.persistentObserver import PersistentPollingEmitter, PersistentPollingObserver, ObservedWatch, DirectorySnapshot, EmptyDirectorySnapshot
from persistentObserver.folderUpdateHandler import FolderUpdateHandler
import logging


class MockEventHandler:
    def __init__(self):
        self.recentSnapshot = None
        self.currentSnapshot = None

    def update_snapshot(self, snapshot):
        self.recentSnapshot = self.currentSnapshot
        self.currentSnapshot = snapshot

    def get_snapshot(self):
        return self.currentSnapshot


class MockScanner:

    def scan_for_photos(self, path):
        return path


@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir


@pytest.fixture
def temp_file():
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        yield tmpfile.name
    os.remove(tmpfile.name)


def test_emitter_initialization(temp_dir, temp_file):
    event_queue = []
    watch = ObservedWatch(temp_dir, recursive=True)
    event_handler = MockEventHandler()

    emitter = PersistentPollingEmitter(event_queue, watch, event_handler, temp_file)
    assert isinstance(emitter._snapshot, EmptyDirectorySnapshot)


def test_emitter_snapshot_handling(temp_dir, temp_file):
    event_queue = []
    watch = ObservedWatch(temp_dir, recursive=True)
    event_handler = MockEventHandler()

    # Create a snapshot file
    snapshot = DirectorySnapshot(temp_dir, recursive=True)
    with open(temp_file, 'wb') as f:
        pickle.dump({temp_dir: snapshot}, f)

    emitter = PersistentPollingEmitter(event_queue, watch, event_handler, temp_file)
    emitter.on_thread_start()

    assert emitter._hasPreviousSnapshot
    assert emitter._snapshot.paths == snapshot.paths
    assert emitter._snapshot.stat_info(temp_dir) == snapshot.stat_info(temp_dir)


def test_emitter_queue_events(temp_dir, temp_file):
    event_queue = EventQueue()
    watch = ObservedWatch(temp_dir, recursive=True)
    event_handler = MockEventHandler()

    emitter = PersistentPollingEmitter(event_queue, watch, event_handler, temp_file)
    emitter.on_thread_start()

    # Create a new file to trigger an event
    new_file_path = os.path.join(temp_dir, 'new_file.txt')
    with open(new_file_path, 'w') as f:
        f.write('test')

    emitter.queue_events(1)
    assert any(isinstance(event[0], FileCreatedEvent) for event in event_queue.queue)


def test_observer_initialization(temp_file):
    observer = PersistentPollingObserver(save_to=temp_file)
    assert observer._filename == temp_file
    assert observer._protocol == 0


def test_observer_schedule(temp_dir, temp_file):
    observer = PersistentPollingObserver(save_to=temp_file)
    event_handler = MockEventHandler()

    watch = observer.schedule(event_handler, temp_dir, recursive=True)
    assert watch.path == temp_dir
    assert watch.is_recursive


def test_emitter_error_handling(temp_dir, temp_file):
    event_queue = EventQueue()
    watch = ObservedWatch(temp_dir, recursive=True)
    event_handler = MockEventHandler()

    # Simulate a permission error by making the directory read-only
    os.chmod(temp_dir, 0o400)
    emitter = PersistentPollingEmitter(event_queue, watch, event_handler, temp_file)
    emitter.on_thread_start()

    try:
        emitter.queue_events(1)
    except OSError:
        assert True  # Expected behavior
    finally:
        os.chmod(temp_dir, 0o700)  # Restore permissions


def test_emitter_concurrency(temp_dir, temp_file):
    event_queue = EventQueue()
    watch = ObservedWatch(temp_dir, recursive=True)
    event_handler = MockEventHandler()

    emitter = PersistentPollingEmitter(event_queue, watch, event_handler, temp_file)
    emitter.on_thread_start()

    def create_file():
        new_file_path = os.path.join(temp_dir, 'concurrent_file.txt')
        with open(new_file_path, 'w') as f:
            f.write('test')
            
    def create_file2():
        new_file_path = os.path.join(temp_dir, 'concurrent_file2.txt')
        with open(new_file_path, 'w') as f:
            f.write('test')

    threads = [threading.Thread(target=create_file) for _ in range(10)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    emitter.queue_events(0)
    assert len(event_queue.queue) <= 2 # sometimes logs a created file as well as a modified directory.
    
    threads = [threading.Thread(target=create_file) for _ in range(10)]
    threads.append(threading.Thread(target=create_file2))
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    emitter.queue_events(0)
    print(event_queue.queue)
    assert len(event_queue.queue) >= 2


def test_observer_persistence(temp_dir, temp_file, caplog):
    with caplog.at_level(logging.DEBUG):
        observer = PersistentPollingObserver(save_to=temp_file)
        event_handler = FolderUpdateHandler([], MockScanner)

        event_snapshot = event_handler.currentSnapshot.path('')

        assert event_snapshot is None

        observer.schedule(event_handler, temp_dir, recursive=True)
        observer.start()
        observer.stop()
        observer.join()
        snapshot = event_handler.currentSnapshot

        # snapshots = {emitter.watch.path: emitter.get_handler_snapshot() for emitter in observer.emitters}
        # with open('data.pkl', 'wb') as file:
        #     pickle.dump(snapshots, file)

        # Load data from the file
        with open(temp_file, 'rb') as file:
            loaded_data = pickle.load(file)

        assert pickle.dumps(loaded_data[temp_dir]) == pickle.dumps(snapshot)
        # Restart observer and check if it loads the previous snapshot
        new_observer = PersistentPollingObserver(save_to=temp_file)
        new_event_handler = FolderUpdateHandler([], MockScanner)
        new_observer.schedule(new_event_handler, temp_dir, recursive=True)
        new_observer.start()

        # Wait a bit to allow the observer to pick up changes
        time.sleep(1)
        new_snapshot = event_handler.currentSnapshot
        assert new_snapshot == snapshot
        paths_list = list(new_event_handler.currentSnapshot.paths)

        assert paths_list[0] == temp_dir

        assert new_event_handler.currentSnapshot.path(temp_dir) is None
